import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:stack_trace/stack_trace.dart';

/// The plugin's façade - Use this class to configure CrashOps SDK.
class CrashOps {
  static const STATUS_OK = "OK";

  /// Returns the instance of CrashOps plugin
  factory CrashOps() {
    return _singleton;
  }

  static const MethodChannel _channel = const MethodChannel('crashops_flutter');

  /// Determines whether CrashOps will / won't collect crashes / errors on debug mode. The default value is `false`.
  bool isEnabledInDebugMode = false;

  /// Determines whether CrashOps will / won't operate, useful if you wish to remotely set it off. The default value is `true`.
  Future<bool> get isEnabled async {
    if (_isEnabled != null) return _isEnabled;

    _isEnabled = await _channel.invokeMethod('CrashOps#isEnabled');

    return _isEnabled;
  }

  /// Sets the CrashOps behaviour, useful if you wish to remotely set it off. The default value is `true`.
  Future<bool> setEnabled(bool value) async {
    var result = await _channel.invokeMethod('CrashOps#setEnabled', value);
    bool didUpdate = result == STATUS_OK;
    if (didUpdate) {
      _isEnabled = value;
    }

    return didUpdate;
  }

  /// You may optionally attach more data if you like, this metadata will be included in all future reports.
  Future<bool> setMetadata(Map<String, dynamic> metadata) async {
    var result = await _channel.invokeMethod('CrashOps#setMetadata', metadata);
    return result == STATUS_OK;
  }

  /// Sets the client ID (recommended). Mandatory only if you wish to store your logs in CrashOps servers.
  Future<bool> setClientId(String clientId) async {
    var result = await _channel.invokeMethod('CrashOps#setClientId', clientId);
    return result == STATUS_OK;
  }

  /// Programmatically logs a non-fatal error, this error also sent to CrashOps servers in case that `clientId` exists.
  Future<bool> onError(dynamic error, [StackTrace stackTrace]) async {
    if (error == null) return false;

    bool inDebugMode = false;
    if (!isEnabledInDebugMode) {
      assert(inDebugMode = true);
    }

    String result;

    FlutterErrorDetails errorDetails;
    if (error is FlutterErrorDetails) {
      errorDetails = error;
    } else {
      errorDetails =
          FlutterErrorDetails(exception: error, stack: stackTrace);
    }

    if (inDebugMode && !isEnabledInDebugMode) {
      print(errorDetails);
      result = STATUS_OK;
    } else {
      final List<String> stackTraceLines = Trace.format(errorDetails.stack)
          .trimRight()
          .split('\n');

      // Yes, we learned from the best:
      // https://pub.dev/documentation/firebase_crashlytics/latest/firebase_crashlytics/Crashlytics/getStackTraceElements.html
      final List<Map<String, String>> stackTraceElements = _getStackTraceElements(
          stackTraceLines);

      result = await _channel
          .invokeMethod('CrashOps#onFlutterError', <String, dynamic>{
        'exception': errorDetails.exceptionAsString() ?? "none",
        'context': '${errorDetails.context ?? "none"}',
        'stackTraceElements': stackTraceElements,
      });
    }

    return result == STATUS_OK;
  }

  /*  Private methods  */

  bool _isEnabled;

  static final CrashOps _singleton = new CrashOps._internal();

  CrashOps._internal() {
    _isEnabled = null;
  }

  static List<Map<String, String>> _getStackTraceElements(List<String> lines) {
    final List<Map<String, String>> elements = <Map<String, String>>[];
    for (String line in lines) {
      final List<String> lineParts = line.split(RegExp('\\s+'));
      try {
        final String fileName = lineParts[0];
        final String lineNumber = lineParts[1].contains(":")
            ? lineParts[1].substring(0, lineParts[1].indexOf(":")).trim()
            : lineParts[1];

        final Map<String, String> element = <String, String>{
          'file': fileName,
          'line': lineNumber,
        };

        if (lineParts[2].contains(".")) {
          final String className =
              lineParts[2].substring(0, lineParts[2].indexOf(".")).trim();
          final String methodName =
              lineParts[2].substring(lineParts[2].indexOf(".") + 1).trim();

          element['class'] = className;
          element['method'] = methodName;
        } else {
          element['method'] = lineParts[2];
        }

        elements.add(element);
      } catch (e) {
        print(e.toString());
      }
    }
    return elements;
  }
}
